/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "accounts-user-generated.h"

#include <string.h>
#ifdef G_OS_UNIX
#  include <gio/gunixfdlist.h>
#endif

#ifdef G_ENABLE_DEBUG
#define g_marshal_value_peek_boolean(v)  g_value_get_boolean (v)
#define g_marshal_value_peek_char(v)     g_value_get_schar (v)
#define g_marshal_value_peek_uchar(v)    g_value_get_uchar (v)
#define g_marshal_value_peek_int(v)      g_value_get_int (v)
#define g_marshal_value_peek_uint(v)     g_value_get_uint (v)
#define g_marshal_value_peek_long(v)     g_value_get_long (v)
#define g_marshal_value_peek_ulong(v)    g_value_get_ulong (v)
#define g_marshal_value_peek_int64(v)    g_value_get_int64 (v)
#define g_marshal_value_peek_uint64(v)   g_value_get_uint64 (v)
#define g_marshal_value_peek_enum(v)     g_value_get_enum (v)
#define g_marshal_value_peek_flags(v)    g_value_get_flags (v)
#define g_marshal_value_peek_float(v)    g_value_get_float (v)
#define g_marshal_value_peek_double(v)   g_value_get_double (v)
#define g_marshal_value_peek_string(v)   (char*) g_value_get_string (v)
#define g_marshal_value_peek_param(v)    g_value_get_param (v)
#define g_marshal_value_peek_boxed(v)    g_value_get_boxed (v)
#define g_marshal_value_peek_pointer(v)  g_value_get_pointer (v)
#define g_marshal_value_peek_object(v)   g_value_get_object (v)
#define g_marshal_value_peek_variant(v)  g_value_get_variant (v)
#else /* !G_ENABLE_DEBUG */
/* WARNING: This code accesses GValues directly, which is UNSUPPORTED API.
 *          Do not access GValues directly in your code. Instead, use the
 *          g_value_get_*() functions
 */
#define g_marshal_value_peek_boolean(v)  (v)->data[0].v_int
#define g_marshal_value_peek_char(v)     (v)->data[0].v_int
#define g_marshal_value_peek_uchar(v)    (v)->data[0].v_uint
#define g_marshal_value_peek_int(v)      (v)->data[0].v_int
#define g_marshal_value_peek_uint(v)     (v)->data[0].v_uint
#define g_marshal_value_peek_long(v)     (v)->data[0].v_long
#define g_marshal_value_peek_ulong(v)    (v)->data[0].v_ulong
#define g_marshal_value_peek_int64(v)    (v)->data[0].v_int64
#define g_marshal_value_peek_uint64(v)   (v)->data[0].v_uint64
#define g_marshal_value_peek_enum(v)     (v)->data[0].v_long
#define g_marshal_value_peek_flags(v)    (v)->data[0].v_ulong
#define g_marshal_value_peek_float(v)    (v)->data[0].v_float
#define g_marshal_value_peek_double(v)   (v)->data[0].v_double
#define g_marshal_value_peek_string(v)   (v)->data[0].v_pointer
#define g_marshal_value_peek_param(v)    (v)->data[0].v_pointer
#define g_marshal_value_peek_boxed(v)    (v)->data[0].v_pointer
#define g_marshal_value_peek_pointer(v)  (v)->data[0].v_pointer
#define g_marshal_value_peek_object(v)   (v)->data[0].v_pointer
#define g_marshal_value_peek_variant(v)  (v)->data[0].v_pointer
#endif /* !G_ENABLE_DEBUG */

typedef struct
{
  GDBusArgInfo parent_struct;
  gboolean use_gvariant;
} _ExtendedGDBusArgInfo;

typedef struct
{
  GDBusMethodInfo parent_struct;
  const gchar *signal_name;
  gboolean pass_fdlist;
} _ExtendedGDBusMethodInfo;

typedef struct
{
  GDBusSignalInfo parent_struct;
  const gchar *signal_name;
} _ExtendedGDBusSignalInfo;

typedef struct
{
  GDBusPropertyInfo parent_struct;
  const gchar *hyphen_name;
  guint use_gvariant : 1;
  guint emits_changed_signal : 1;
} _ExtendedGDBusPropertyInfo;

typedef struct
{
  GDBusInterfaceInfo parent_struct;
  const gchar *hyphen_name;
} _ExtendedGDBusInterfaceInfo;

typedef struct
{
  const _ExtendedGDBusPropertyInfo *info;
  guint prop_id;
  GValue orig_value; /* the value before the change */
} ChangedProperty;

static void
_changed_property_free (ChangedProperty *data)
{
  g_value_unset (&data->orig_value);
  g_free (data);
}

static gboolean
_g_strv_equal0 (gchar **a, gchar **b)
{
  gboolean ret = FALSE;
  guint n;
  if (a == NULL && b == NULL)
    {
      ret = TRUE;
      goto out;
    }
  if (a == NULL || b == NULL)
    goto out;
  if (g_strv_length (a) != g_strv_length (b))
    goto out;
  for (n = 0; a[n] != NULL; n++)
    if (g_strcmp0 (a[n], b[n]) != 0)
      goto out;
  ret = TRUE;
out:
  return ret;
}

static gboolean
_g_variant_equal0 (GVariant *a, GVariant *b)
{
  gboolean ret = FALSE;
  if (a == NULL && b == NULL)
    {
      ret = TRUE;
      goto out;
    }
  if (a == NULL || b == NULL)
    goto out;
  ret = g_variant_equal (a, b);
out:
  return ret;
}

G_GNUC_UNUSED static gboolean
_g_value_equal (const GValue *a, const GValue *b)
{
  gboolean ret = FALSE;
  g_assert (G_VALUE_TYPE (a) == G_VALUE_TYPE (b));
  switch (G_VALUE_TYPE (a))
    {
      case G_TYPE_BOOLEAN:
        ret = (g_value_get_boolean (a) == g_value_get_boolean (b));
        break;
      case G_TYPE_UCHAR:
        ret = (g_value_get_uchar (a) == g_value_get_uchar (b));
        break;
      case G_TYPE_INT:
        ret = (g_value_get_int (a) == g_value_get_int (b));
        break;
      case G_TYPE_UINT:
        ret = (g_value_get_uint (a) == g_value_get_uint (b));
        break;
      case G_TYPE_INT64:
        ret = (g_value_get_int64 (a) == g_value_get_int64 (b));
        break;
      case G_TYPE_UINT64:
        ret = (g_value_get_uint64 (a) == g_value_get_uint64 (b));
        break;
      case G_TYPE_DOUBLE:
        {
          /* Avoid -Wfloat-equal warnings by doing a direct bit compare */
          gdouble da = g_value_get_double (a);
          gdouble db = g_value_get_double (b);
          ret = memcmp (&da, &db, sizeof (gdouble)) == 0;
        }
        break;
      case G_TYPE_STRING:
        ret = (g_strcmp0 (g_value_get_string (a), g_value_get_string (b)) == 0);
        break;
      case G_TYPE_VARIANT:
        ret = _g_variant_equal0 (g_value_get_variant (a), g_value_get_variant (b));
        break;
      default:
        if (G_VALUE_TYPE (a) == G_TYPE_STRV)
          ret = _g_strv_equal0 (g_value_get_boxed (a), g_value_get_boxed (b));
        else
          g_critical ("_g_value_equal() does not handle type %s", g_type_name (G_VALUE_TYPE (a)));
        break;
    }
  return ret;
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_name,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOXED (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectBoxedFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *const *arg_languages,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectBoxedFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectBoxedFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_boxed (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectBooleanFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gboolean arg_locked,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectBooleanFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectBooleanFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_boolean (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectIntFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint arg_accountType,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectIntFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectIntFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringStringFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_password,
        const gchar *arg_hint,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringStringFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringStringFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              g_marshal_value_peek_string (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 2);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT64_INT64_INT64_INT64 (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectInt64Int64Int64Int64Func)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint64 arg_min_days_between_changes,
        gint64 arg_max_days_between_changes,
        gint64 arg_days_to_warn,
        gint64 arg_days_after_expiration_until_lock,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectInt64Int64Int64Int64Func callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 6);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectInt64Int64Int64Int64Func)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int64 (param_values + 2),
              g_marshal_value_peek_int64 (param_values + 3),
              g_marshal_value_peek_int64 (param_values + 4),
              g_marshal_value_peek_int64 (param_values + 5),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT64 (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectInt64Func)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint64 arg_expiration_time,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectInt64Func callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectInt64Func)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int64 (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

/* ------------------------------------------------------------------------
 * Code for interface org.freedesktop.Accounts.User
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:AccountsUser
 * @title: AccountsUser
 * @short_description: Generated C code for the org.freedesktop.Accounts.User D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link> D-Bus interface in C.
 */

enum
{
  ACCOUNTS__USER_CHANGED,
};

static unsigned ACCOUNTS__USER_SIGNALS[1] = { 0 };

/* ---- Introspection data for org.freedesktop.Accounts.User ---- */

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_user_name_IN_ARG_name =
{
  {
    -1,
    (gchar *) "name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_user_name_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_user_name_IN_ARG_name.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_user_name_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_user_name_annotation_info_pointers[] =
{
  &_accounts_user_method_set_user_name_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_user_name =
{
  {
    -1,
    (gchar *) "SetUserName",
    (GDBusArgInfo **) &_accounts_user_method_info_set_user_name_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_user_name_annotation_info_pointers
  },
  "handle-set-user-name",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_real_name_IN_ARG_name =
{
  {
    -1,
    (gchar *) "name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_real_name_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_real_name_IN_ARG_name.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_real_name_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_real_name_annotation_info_pointers[] =
{
  &_accounts_user_method_set_real_name_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_real_name =
{
  {
    -1,
    (gchar *) "SetRealName",
    (GDBusArgInfo **) &_accounts_user_method_info_set_real_name_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_real_name_annotation_info_pointers
  },
  "handle-set-real-name",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_email_IN_ARG_email =
{
  {
    -1,
    (gchar *) "email",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_email_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_email_IN_ARG_email.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_email_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_email_annotation_info_pointers[] =
{
  &_accounts_user_method_set_email_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_email =
{
  {
    -1,
    (gchar *) "SetEmail",
    (GDBusArgInfo **) &_accounts_user_method_info_set_email_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_email_annotation_info_pointers
  },
  "handle-set-email",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_language_IN_ARG_language =
{
  {
    -1,
    (gchar *) "language",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_language_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_language_IN_ARG_language.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_language_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_language_annotation_info_pointers[] =
{
  &_accounts_user_method_set_language_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_language =
{
  {
    -1,
    (gchar *) "SetLanguage",
    (GDBusArgInfo **) &_accounts_user_method_info_set_language_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_language_annotation_info_pointers
  },
  "handle-set-language",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_languages_IN_ARG_languages =
{
  {
    -1,
    (gchar *) "languages",
    (gchar *) "as",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_languages_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_languages_IN_ARG_languages.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_languages_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_languages_annotation_info_pointers[] =
{
  &_accounts_user_method_set_languages_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_languages =
{
  {
    -1,
    (gchar *) "SetLanguages",
    (GDBusArgInfo **) &_accounts_user_method_info_set_languages_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_languages_annotation_info_pointers
  },
  "handle-set-languages",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_xsession_IN_ARG_x_session =
{
  {
    -1,
    (gchar *) "x_session",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_xsession_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_xsession_IN_ARG_x_session.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_xsession_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_xsession_annotation_info_1 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.CSymbol",
  (gchar *) "user_set_x_session",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_xsession_annotation_info_pointers[] =
{
  &_accounts_user_method_set_xsession_annotation_info_0,
  &_accounts_user_method_set_xsession_annotation_info_1,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_xsession =
{
  {
    -1,
    (gchar *) "SetXSession",
    (GDBusArgInfo **) &_accounts_user_method_info_set_xsession_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_xsession_annotation_info_pointers
  },
  "handle-set-xsession",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_session_IN_ARG_session =
{
  {
    -1,
    (gchar *) "session",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_session_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_session_IN_ARG_session.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_session_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_session_annotation_info_1 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.CSymbol",
  (gchar *) "user_set_session",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_session_annotation_info_pointers[] =
{
  &_accounts_user_method_set_session_annotation_info_0,
  &_accounts_user_method_set_session_annotation_info_1,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_session =
{
  {
    -1,
    (gchar *) "SetSession",
    (GDBusArgInfo **) &_accounts_user_method_info_set_session_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_session_annotation_info_pointers
  },
  "handle-set-session",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_session_type_IN_ARG_session_type =
{
  {
    -1,
    (gchar *) "session_type",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_session_type_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_session_type_IN_ARG_session_type.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_session_type_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_session_type_annotation_info_1 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.CSymbol",
  (gchar *) "user_set_session_type",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_session_type_annotation_info_pointers[] =
{
  &_accounts_user_method_set_session_type_annotation_info_0,
  &_accounts_user_method_set_session_type_annotation_info_1,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_session_type =
{
  {
    -1,
    (gchar *) "SetSessionType",
    (GDBusArgInfo **) &_accounts_user_method_info_set_session_type_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_session_type_annotation_info_pointers
  },
  "handle-set-session-type",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_location_IN_ARG_location =
{
  {
    -1,
    (gchar *) "location",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_location_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_location_IN_ARG_location.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_location_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_location_annotation_info_pointers[] =
{
  &_accounts_user_method_set_location_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_location =
{
  {
    -1,
    (gchar *) "SetLocation",
    (GDBusArgInfo **) &_accounts_user_method_info_set_location_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_location_annotation_info_pointers
  },
  "handle-set-location",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_home_directory_IN_ARG_homedir =
{
  {
    -1,
    (gchar *) "homedir",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_home_directory_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_home_directory_IN_ARG_homedir.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_home_directory_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_home_directory_annotation_info_pointers[] =
{
  &_accounts_user_method_set_home_directory_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_home_directory =
{
  {
    -1,
    (gchar *) "SetHomeDirectory",
    (GDBusArgInfo **) &_accounts_user_method_info_set_home_directory_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_home_directory_annotation_info_pointers
  },
  "handle-set-home-directory",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_shell_IN_ARG_shell =
{
  {
    -1,
    (gchar *) "shell",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_shell_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_shell_IN_ARG_shell.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_shell_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_shell_annotation_info_pointers[] =
{
  &_accounts_user_method_set_shell_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_shell =
{
  {
    -1,
    (gchar *) "SetShell",
    (GDBusArgInfo **) &_accounts_user_method_info_set_shell_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_shell_annotation_info_pointers
  },
  "handle-set-shell",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_icon_file_IN_ARG_filename =
{
  {
    -1,
    (gchar *) "filename",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_icon_file_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_icon_file_IN_ARG_filename.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_icon_file_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_icon_file_annotation_info_pointers[] =
{
  &_accounts_user_method_set_icon_file_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_icon_file =
{
  {
    -1,
    (gchar *) "SetIconFile",
    (GDBusArgInfo **) &_accounts_user_method_info_set_icon_file_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_icon_file_annotation_info_pointers
  },
  "handle-set-icon-file",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_locked_IN_ARG_locked =
{
  {
    -1,
    (gchar *) "locked",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_locked_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_locked_IN_ARG_locked.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_locked_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_locked_annotation_info_pointers[] =
{
  &_accounts_user_method_set_locked_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_locked =
{
  {
    -1,
    (gchar *) "SetLocked",
    (GDBusArgInfo **) &_accounts_user_method_info_set_locked_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_locked_annotation_info_pointers
  },
  "handle-set-locked",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_account_type_IN_ARG_accountType =
{
  {
    -1,
    (gchar *) "accountType",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_account_type_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_account_type_IN_ARG_accountType.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_account_type_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_account_type_annotation_info_pointers[] =
{
  &_accounts_user_method_set_account_type_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_account_type =
{
  {
    -1,
    (gchar *) "SetAccountType",
    (GDBusArgInfo **) &_accounts_user_method_info_set_account_type_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_account_type_annotation_info_pointers
  },
  "handle-set-account-type",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_password_mode_IN_ARG_mode =
{
  {
    -1,
    (gchar *) "mode",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_password_mode_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_password_mode_IN_ARG_mode.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_password_mode_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_password_mode_annotation_info_pointers[] =
{
  &_accounts_user_method_set_password_mode_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_password_mode =
{
  {
    -1,
    (gchar *) "SetPasswordMode",
    (GDBusArgInfo **) &_accounts_user_method_info_set_password_mode_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_password_mode_annotation_info_pointers
  },
  "handle-set-password-mode",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_password_IN_ARG_password =
{
  {
    -1,
    (gchar *) "password",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_password_IN_ARG_hint =
{
  {
    -1,
    (gchar *) "hint",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_password_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_password_IN_ARG_password.parent_struct,
  &_accounts_user_method_info_set_password_IN_ARG_hint.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_password_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_password_annotation_info_pointers[] =
{
  &_accounts_user_method_set_password_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_password =
{
  {
    -1,
    (gchar *) "SetPassword",
    (GDBusArgInfo **) &_accounts_user_method_info_set_password_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_password_annotation_info_pointers
  },
  "handle-set-password",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_password_hint_IN_ARG_hint =
{
  {
    -1,
    (gchar *) "hint",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_password_hint_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_password_hint_IN_ARG_hint.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_password_hint_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_password_hint_annotation_info_pointers[] =
{
  &_accounts_user_method_set_password_hint_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_password_hint =
{
  {
    -1,
    (gchar *) "SetPasswordHint",
    (GDBusArgInfo **) &_accounts_user_method_info_set_password_hint_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_password_hint_annotation_info_pointers
  },
  "handle-set-password-hint",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_automatic_login_IN_ARG_enabled =
{
  {
    -1,
    (gchar *) "enabled",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_automatic_login_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_automatic_login_IN_ARG_enabled.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_automatic_login_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_automatic_login_annotation_info_pointers[] =
{
  &_accounts_user_method_set_automatic_login_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_automatic_login =
{
  {
    -1,
    (gchar *) "SetAutomaticLogin",
    (GDBusArgInfo **) &_accounts_user_method_info_set_automatic_login_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_automatic_login_annotation_info_pointers
  },
  "handle-set-automatic-login",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_get_password_expiration_policy_OUT_ARG_expiration_time =
{
  {
    -1,
    (gchar *) "expiration_time",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_get_password_expiration_policy_OUT_ARG_last_change_time =
{
  {
    -1,
    (gchar *) "last_change_time",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_get_password_expiration_policy_OUT_ARG_min_days_between_changes =
{
  {
    -1,
    (gchar *) "min_days_between_changes",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_get_password_expiration_policy_OUT_ARG_max_days_between_changes =
{
  {
    -1,
    (gchar *) "max_days_between_changes",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_get_password_expiration_policy_OUT_ARG_days_to_warn =
{
  {
    -1,
    (gchar *) "days_to_warn",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_get_password_expiration_policy_OUT_ARG_days_after_expiration_until_lock =
{
  {
    -1,
    (gchar *) "days_after_expiration_until_lock",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_get_password_expiration_policy_OUT_ARG_pointers[] =
{
  &_accounts_user_method_info_get_password_expiration_policy_OUT_ARG_expiration_time.parent_struct,
  &_accounts_user_method_info_get_password_expiration_policy_OUT_ARG_last_change_time.parent_struct,
  &_accounts_user_method_info_get_password_expiration_policy_OUT_ARG_min_days_between_changes.parent_struct,
  &_accounts_user_method_info_get_password_expiration_policy_OUT_ARG_max_days_between_changes.parent_struct,
  &_accounts_user_method_info_get_password_expiration_policy_OUT_ARG_days_to_warn.parent_struct,
  &_accounts_user_method_info_get_password_expiration_policy_OUT_ARG_days_after_expiration_until_lock.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_get_password_expiration_policy_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_get_password_expiration_policy_annotation_info_pointers[] =
{
  &_accounts_user_method_get_password_expiration_policy_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_get_password_expiration_policy =
{
  {
    -1,
    (gchar *) "GetPasswordExpirationPolicy",
    NULL,
    (GDBusArgInfo **) &_accounts_user_method_info_get_password_expiration_policy_OUT_ARG_pointers,
    (GDBusAnnotationInfo **) &_accounts_user_method_get_password_expiration_policy_annotation_info_pointers
  },
  "handle-get-password-expiration-policy",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_password_expiration_policy_IN_ARG_min_days_between_changes =
{
  {
    -1,
    (gchar *) "min_days_between_changes",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_password_expiration_policy_IN_ARG_max_days_between_changes =
{
  {
    -1,
    (gchar *) "max_days_between_changes",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_password_expiration_policy_IN_ARG_days_to_warn =
{
  {
    -1,
    (gchar *) "days_to_warn",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_password_expiration_policy_IN_ARG_days_after_expiration_until_lock =
{
  {
    -1,
    (gchar *) "days_after_expiration_until_lock",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_password_expiration_policy_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_password_expiration_policy_IN_ARG_min_days_between_changes.parent_struct,
  &_accounts_user_method_info_set_password_expiration_policy_IN_ARG_max_days_between_changes.parent_struct,
  &_accounts_user_method_info_set_password_expiration_policy_IN_ARG_days_to_warn.parent_struct,
  &_accounts_user_method_info_set_password_expiration_policy_IN_ARG_days_after_expiration_until_lock.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_password_expiration_policy_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_password_expiration_policy_annotation_info_pointers[] =
{
  &_accounts_user_method_set_password_expiration_policy_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_password_expiration_policy =
{
  {
    -1,
    (gchar *) "SetPasswordExpirationPolicy",
    (GDBusArgInfo **) &_accounts_user_method_info_set_password_expiration_policy_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_password_expiration_policy_annotation_info_pointers
  },
  "handle-set-password-expiration-policy",
  FALSE
};

static const _ExtendedGDBusArgInfo _accounts_user_method_info_set_user_expiration_policy_IN_ARG_expiration_time =
{
  {
    -1,
    (gchar *) "expiration_time",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _accounts_user_method_info_set_user_expiration_policy_IN_ARG_pointers[] =
{
  &_accounts_user_method_info_set_user_expiration_policy_IN_ARG_expiration_time.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _accounts_user_method_set_user_expiration_policy_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_method_set_user_expiration_policy_annotation_info_pointers[] =
{
  &_accounts_user_method_set_user_expiration_policy_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _accounts_user_method_info_set_user_expiration_policy =
{
  {
    -1,
    (gchar *) "SetUserExpirationPolicy",
    (GDBusArgInfo **) &_accounts_user_method_info_set_user_expiration_policy_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_accounts_user_method_set_user_expiration_policy_annotation_info_pointers
  },
  "handle-set-user-expiration-policy",
  FALSE
};

static const GDBusMethodInfo * const _accounts_user_method_info_pointers[] =
{
  &_accounts_user_method_info_set_user_name.parent_struct,
  &_accounts_user_method_info_set_real_name.parent_struct,
  &_accounts_user_method_info_set_email.parent_struct,
  &_accounts_user_method_info_set_language.parent_struct,
  &_accounts_user_method_info_set_languages.parent_struct,
  &_accounts_user_method_info_set_xsession.parent_struct,
  &_accounts_user_method_info_set_session.parent_struct,
  &_accounts_user_method_info_set_session_type.parent_struct,
  &_accounts_user_method_info_set_location.parent_struct,
  &_accounts_user_method_info_set_home_directory.parent_struct,
  &_accounts_user_method_info_set_shell.parent_struct,
  &_accounts_user_method_info_set_icon_file.parent_struct,
  &_accounts_user_method_info_set_locked.parent_struct,
  &_accounts_user_method_info_set_account_type.parent_struct,
  &_accounts_user_method_info_set_password_mode.parent_struct,
  &_accounts_user_method_info_set_password.parent_struct,
  &_accounts_user_method_info_set_password_hint.parent_struct,
  &_accounts_user_method_info_set_automatic_login.parent_struct,
  &_accounts_user_method_info_get_password_expiration_policy.parent_struct,
  &_accounts_user_method_info_set_password_expiration_policy.parent_struct,
  &_accounts_user_method_info_set_user_expiration_policy.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _accounts_user_signal_info_changed =
{
  {
    -1,
    (gchar *) "Changed",
    NULL,
    NULL
  },
  "changed"
};

static const GDBusSignalInfo * const _accounts_user_signal_info_pointers[] =
{
  &_accounts_user_signal_info_changed.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_uid =
{
  {
    -1,
    (gchar *) "Uid",
    (gchar *) "t",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "uid",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_user_name =
{
  {
    -1,
    (gchar *) "UserName",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "user-name",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_real_name =
{
  {
    -1,
    (gchar *) "RealName",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "real-name",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_account_type =
{
  {
    -1,
    (gchar *) "AccountType",
    (gchar *) "i",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "account-type",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_home_directory =
{
  {
    -1,
    (gchar *) "HomeDirectory",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "home-directory",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_shell =
{
  {
    -1,
    (gchar *) "Shell",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "shell",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_email =
{
  {
    -1,
    (gchar *) "Email",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "email",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_language =
{
  {
    -1,
    (gchar *) "Language",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "language",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_languages =
{
  {
    -1,
    (gchar *) "Languages",
    (gchar *) "as",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "languages",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_session =
{
  {
    -1,
    (gchar *) "Session",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "session",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_session_type =
{
  {
    -1,
    (gchar *) "SessionType",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "session-type",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_xsession =
{
  {
    -1,
    (gchar *) "XSession",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "xsession",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_location =
{
  {
    -1,
    (gchar *) "Location",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "location",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_login_frequency =
{
  {
    -1,
    (gchar *) "LoginFrequency",
    (gchar *) "t",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "login-frequency",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_login_time =
{
  {
    -1,
    (gchar *) "LoginTime",
    (gchar *) "x",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "login-time",
  FALSE,
  TRUE
};

static const GDBusAnnotationInfo _accounts_user_property_login_history_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.Property.EmitsChangedSignal",
  (gchar *) "false",
  NULL
};

static const GDBusAnnotationInfo * const _accounts_user_property_login_history_annotation_info_pointers[] =
{
  &_accounts_user_property_login_history_annotation_info_0,
  NULL
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_login_history =
{
  {
    -1,
    (gchar *) "LoginHistory",
    (gchar *) "a(xxa{sv})",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    (GDBusAnnotationInfo **) &_accounts_user_property_login_history_annotation_info_pointers
  },
  "login-history",
  FALSE,
  FALSE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_icon_file =
{
  {
    -1,
    (gchar *) "IconFile",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "icon-file",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_saved =
{
  {
    -1,
    (gchar *) "Saved",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "saved",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_locked =
{
  {
    -1,
    (gchar *) "Locked",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "locked",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_password_mode =
{
  {
    -1,
    (gchar *) "PasswordMode",
    (gchar *) "i",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "password-mode",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_password_hint =
{
  {
    -1,
    (gchar *) "PasswordHint",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "password-hint",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_automatic_login =
{
  {
    -1,
    (gchar *) "AutomaticLogin",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "automatic-login",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_system_account =
{
  {
    -1,
    (gchar *) "SystemAccount",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "system-account",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_local_account =
{
  {
    -1,
    (gchar *) "LocalAccount",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "local-account",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _accounts_user_property_info_uses_homed =
{
  {
    -1,
    (gchar *) "UsesHomed",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "uses-homed",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _accounts_user_property_info_pointers[] =
{
  &_accounts_user_property_info_uid.parent_struct,
  &_accounts_user_property_info_user_name.parent_struct,
  &_accounts_user_property_info_real_name.parent_struct,
  &_accounts_user_property_info_account_type.parent_struct,
  &_accounts_user_property_info_home_directory.parent_struct,
  &_accounts_user_property_info_shell.parent_struct,
  &_accounts_user_property_info_email.parent_struct,
  &_accounts_user_property_info_language.parent_struct,
  &_accounts_user_property_info_languages.parent_struct,
  &_accounts_user_property_info_session.parent_struct,
  &_accounts_user_property_info_session_type.parent_struct,
  &_accounts_user_property_info_xsession.parent_struct,
  &_accounts_user_property_info_location.parent_struct,
  &_accounts_user_property_info_login_frequency.parent_struct,
  &_accounts_user_property_info_login_time.parent_struct,
  &_accounts_user_property_info_login_history.parent_struct,
  &_accounts_user_property_info_icon_file.parent_struct,
  &_accounts_user_property_info_saved.parent_struct,
  &_accounts_user_property_info_locked.parent_struct,
  &_accounts_user_property_info_password_mode.parent_struct,
  &_accounts_user_property_info_password_hint.parent_struct,
  &_accounts_user_property_info_automatic_login.parent_struct,
  &_accounts_user_property_info_system_account.parent_struct,
  &_accounts_user_property_info_local_account.parent_struct,
  &_accounts_user_property_info_uses_homed.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _accounts_user_interface_info =
{
  {
    -1,
    (gchar *) "org.freedesktop.Accounts.User",
    (GDBusMethodInfo **) &_accounts_user_method_info_pointers,
    (GDBusSignalInfo **) &_accounts_user_signal_info_pointers,
    (GDBusPropertyInfo **) &_accounts_user_property_info_pointers,
    NULL
  },
  "user",
};


/**
 * accounts_user_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
accounts_user_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_accounts_user_interface_info.parent_struct;
}

/**
 * accounts_user_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #AccountsUser interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
accounts_user_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "uid");
  g_object_class_override_property (klass, property_id_begin++, "user-name");
  g_object_class_override_property (klass, property_id_begin++, "real-name");
  g_object_class_override_property (klass, property_id_begin++, "account-type");
  g_object_class_override_property (klass, property_id_begin++, "home-directory");
  g_object_class_override_property (klass, property_id_begin++, "shell");
  g_object_class_override_property (klass, property_id_begin++, "email");
  g_object_class_override_property (klass, property_id_begin++, "language");
  g_object_class_override_property (klass, property_id_begin++, "languages");
  g_object_class_override_property (klass, property_id_begin++, "session");
  g_object_class_override_property (klass, property_id_begin++, "session-type");
  g_object_class_override_property (klass, property_id_begin++, "xsession");
  g_object_class_override_property (klass, property_id_begin++, "location");
  g_object_class_override_property (klass, property_id_begin++, "login-frequency");
  g_object_class_override_property (klass, property_id_begin++, "login-time");
  g_object_class_override_property (klass, property_id_begin++, "login-history");
  g_object_class_override_property (klass, property_id_begin++, "icon-file");
  g_object_class_override_property (klass, property_id_begin++, "saved");
  g_object_class_override_property (klass, property_id_begin++, "locked");
  g_object_class_override_property (klass, property_id_begin++, "password-mode");
  g_object_class_override_property (klass, property_id_begin++, "password-hint");
  g_object_class_override_property (klass, property_id_begin++, "automatic-login");
  g_object_class_override_property (klass, property_id_begin++, "system-account");
  g_object_class_override_property (klass, property_id_begin++, "local-account");
  g_object_class_override_property (klass, property_id_begin++, "uses-homed");
  return property_id_begin - 1;
}


inline static void
accounts_user_signal_marshal_changed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_user_name (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_real_name (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_email (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_language (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_languages (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOXED (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_xsession (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_session (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_session_type (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_location (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_home_directory (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_shell (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_icon_file (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_locked (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_account_type (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_password_mode (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_password (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_password_hint (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_automatic_login (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_get_password_expiration_policy (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_password_expiration_policy (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT64_INT64_INT64_INT64 (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
accounts_user_method_marshal_set_user_expiration_policy (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT64 (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * AccountsUser:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>.
 */

/**
 * AccountsUserIface:
 * @parent_iface: The parent interface.
 * @handle_get_password_expiration_policy: Handler for the #AccountsUser::handle-get-password-expiration-policy signal.
 * @handle_set_account_type: Handler for the #AccountsUser::handle-set-account-type signal.
 * @handle_set_automatic_login: Handler for the #AccountsUser::handle-set-automatic-login signal.
 * @handle_set_email: Handler for the #AccountsUser::handle-set-email signal.
 * @handle_set_home_directory: Handler for the #AccountsUser::handle-set-home-directory signal.
 * @handle_set_icon_file: Handler for the #AccountsUser::handle-set-icon-file signal.
 * @handle_set_language: Handler for the #AccountsUser::handle-set-language signal.
 * @handle_set_languages: Handler for the #AccountsUser::handle-set-languages signal.
 * @handle_set_location: Handler for the #AccountsUser::handle-set-location signal.
 * @handle_set_locked: Handler for the #AccountsUser::handle-set-locked signal.
 * @handle_set_password: Handler for the #AccountsUser::handle-set-password signal.
 * @handle_set_password_expiration_policy: Handler for the #AccountsUser::handle-set-password-expiration-policy signal.
 * @handle_set_password_hint: Handler for the #AccountsUser::handle-set-password-hint signal.
 * @handle_set_password_mode: Handler for the #AccountsUser::handle-set-password-mode signal.
 * @handle_set_real_name: Handler for the #AccountsUser::handle-set-real-name signal.
 * @handle_set_session: Handler for the #AccountsUser::handle-set-session signal.
 * @handle_set_session_type: Handler for the #AccountsUser::handle-set-session-type signal.
 * @handle_set_shell: Handler for the #AccountsUser::handle-set-shell signal.
 * @handle_set_user_expiration_policy: Handler for the #AccountsUser::handle-set-user-expiration-policy signal.
 * @handle_set_user_name: Handler for the #AccountsUser::handle-set-user-name signal.
 * @handle_set_xsession: Handler for the #AccountsUser::handle-set-xsession signal.
 * @get_account_type: Getter for the #AccountsUser:account-type property.
 * @get_automatic_login: Getter for the #AccountsUser:automatic-login property.
 * @get_email: Getter for the #AccountsUser:email property.
 * @get_home_directory: Getter for the #AccountsUser:home-directory property.
 * @get_icon_file: Getter for the #AccountsUser:icon-file property.
 * @get_language: Getter for the #AccountsUser:language property.
 * @get_languages: Getter for the #AccountsUser:languages property.
 * @get_local_account: Getter for the #AccountsUser:local-account property.
 * @get_location: Getter for the #AccountsUser:location property.
 * @get_locked: Getter for the #AccountsUser:locked property.
 * @get_login_frequency: Getter for the #AccountsUser:login-frequency property.
 * @get_login_history: Getter for the #AccountsUser:login-history property.
 * @get_login_time: Getter for the #AccountsUser:login-time property.
 * @get_password_hint: Getter for the #AccountsUser:password-hint property.
 * @get_password_mode: Getter for the #AccountsUser:password-mode property.
 * @get_real_name: Getter for the #AccountsUser:real-name property.
 * @get_saved: Getter for the #AccountsUser:saved property.
 * @get_session: Getter for the #AccountsUser:session property.
 * @get_session_type: Getter for the #AccountsUser:session-type property.
 * @get_shell: Getter for the #AccountsUser:shell property.
 * @get_system_account: Getter for the #AccountsUser:system-account property.
 * @get_uid: Getter for the #AccountsUser:uid property.
 * @get_user_name: Getter for the #AccountsUser:user-name property.
 * @get_uses_homed: Getter for the #AccountsUser:uses-homed property.
 * @get_xsession: Getter for the #AccountsUser:xsession property.
 * @changed: Handler for the #AccountsUser::changed signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>.
 */

typedef AccountsUserIface AccountsUserInterface;
G_DEFINE_INTERFACE (AccountsUser, accounts_user, G_TYPE_OBJECT)

static void
accounts_user_default_init (AccountsUserIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * AccountsUser::handle-set-user-name:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_name: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserName">SetUserName()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_user_name() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-user-name",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_user_name),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_user_name,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-real-name:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_name: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetRealName">SetRealName()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_real_name() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-real-name",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_real_name),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_real_name,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-email:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_email: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetEmail">SetEmail()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_email() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-email",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_email),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_email,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-language:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_language: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguage">SetLanguage()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_language() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-language",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_language),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_language,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-languages:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_languages: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguages">SetLanguages()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_languages() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-languages",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_languages),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_languages,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRV);

  /**
   * AccountsUser::handle-set-xsession:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_x_session: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetXSession">SetXSession()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_xsession() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-xsession",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_xsession),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_xsession,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-session:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_session: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetSession">SetSession()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_session() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-session",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_session),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_session,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-session-type:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_session_type: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetSessionType">SetSessionType()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_session_type() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-session-type",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_session_type),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_session_type,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-location:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_location: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocation">SetLocation()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_location() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-location",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_location),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_location,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-home-directory:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_homedir: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetHomeDirectory">SetHomeDirectory()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_home_directory() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-home-directory",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_home_directory),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_home_directory,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-shell:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_shell: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetShell">SetShell()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_shell() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-shell",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_shell),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_shell,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-icon-file:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_filename: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetIconFile">SetIconFile()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_icon_file() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-icon-file",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_icon_file),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_icon_file,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-locked:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_locked: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocked">SetLocked()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_locked() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-locked",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_locked),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_locked,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_BOOLEAN);

  /**
   * AccountsUser::handle-set-account-type:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_accountType: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAccountType">SetAccountType()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_account_type() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-account-type",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_account_type),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_account_type,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT);

  /**
   * AccountsUser::handle-set-password-mode:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_mode: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordMode">SetPasswordMode()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_password_mode() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-password-mode",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_password_mode),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_password_mode,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT);

  /**
   * AccountsUser::handle-set-password:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_password: Argument passed by remote caller.
   * @arg_hint: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPassword">SetPassword()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_password() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-password",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_password),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_password,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-password-hint:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_hint: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordHint">SetPasswordHint()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_password_hint() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-password-hint",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_password_hint),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_password_hint,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * AccountsUser::handle-set-automatic-login:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_enabled: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAutomaticLogin">SetAutomaticLogin()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_automatic_login() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-automatic-login",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_automatic_login),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_automatic_login,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_BOOLEAN);

  /**
   * AccountsUser::handle-get-password-expiration-policy:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.GetPasswordExpirationPolicy">GetPasswordExpirationPolicy()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_get_password_expiration_policy() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-get-password-expiration-policy",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_get_password_expiration_policy),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_get_password_expiration_policy,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * AccountsUser::handle-set-password-expiration-policy:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_min_days_between_changes: Argument passed by remote caller.
   * @arg_max_days_between_changes: Argument passed by remote caller.
   * @arg_days_to_warn: Argument passed by remote caller.
   * @arg_days_after_expiration_until_lock: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordExpirationPolicy">SetPasswordExpirationPolicy()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_password_expiration_policy() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-password-expiration-policy",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_password_expiration_policy),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_password_expiration_policy,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT64, G_TYPE_INT64, G_TYPE_INT64, G_TYPE_INT64);

  /**
   * AccountsUser::handle-set-user-expiration-policy:
   * @object: A #AccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_expiration_time: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserExpirationPolicy">SetUserExpirationPolicy()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call accounts_user_complete_set_user_expiration_policy() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-user-expiration-policy",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (AccountsUserIface, handle_set_user_expiration_policy),
    g_signal_accumulator_true_handled,
    NULL,
      accounts_user_method_marshal_set_user_expiration_policy,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT64);

  /* GObject signals for received D-Bus signals: */
  /**
   * AccountsUser::changed:
   * @object: A #AccountsUser.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-freedesktop-Accounts-User.Changed">"Changed"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  ACCOUNTS__USER_SIGNALS[ACCOUNTS__USER_CHANGED] =
    g_signal_new ("changed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (AccountsUserIface, changed),
      NULL,
      NULL,
      accounts_user_signal_marshal_changed,
      G_TYPE_NONE,
      0);

  /* GObject properties for D-Bus properties: */
  /**
   * AccountsUser:uid:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Uid">"Uid"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_uint64 ("uid", "Uid", "Uid", 0, G_MAXUINT64, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:user-name:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.UserName">"UserName"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("user-name", "UserName", "UserName", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:real-name:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.RealName">"RealName"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("real-name", "RealName", "RealName", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:account-type:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.AccountType">"AccountType"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int ("account-type", "AccountType", "AccountType", G_MININT32, G_MAXINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:home-directory:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.HomeDirectory">"HomeDirectory"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("home-directory", "HomeDirectory", "HomeDirectory", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:shell:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Shell">"Shell"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("shell", "Shell", "Shell", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:email:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Email">"Email"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("email", "Email", "Email", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:language:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Language">"Language"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("language", "Language", "Language", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:languages:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Languages">"Languages"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boxed ("languages", "Languages", "Languages", G_TYPE_STRV, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:session:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Session">"Session"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("session", "Session", "Session", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:session-type:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.SessionType">"SessionType"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("session-type", "SessionType", "SessionType", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:xsession:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.XSession">"XSession"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("xsession", "XSession", "XSession", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:location:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Location">"Location"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("location", "Location", "Location", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:login-frequency:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginFrequency">"LoginFrequency"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_uint64 ("login-frequency", "LoginFrequency", "LoginFrequency", 0, G_MAXUINT64, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:login-time:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginTime">"LoginTime"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int64 ("login-time", "LoginTime", "LoginTime", G_MININT64, G_MAXINT64, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:login-history:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginHistory">"LoginHistory"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_variant ("login-history", "LoginHistory", "LoginHistory", G_VARIANT_TYPE ("a(xxa{sv})"), NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:icon-file:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.IconFile">"IconFile"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("icon-file", "IconFile", "IconFile", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:saved:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Saved">"Saved"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("saved", "Saved", "Saved", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:locked:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Locked">"Locked"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("locked", "Locked", "Locked", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:password-mode:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordMode">"PasswordMode"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int ("password-mode", "PasswordMode", "PasswordMode", G_MININT32, G_MAXINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:password-hint:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordHint">"PasswordHint"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("password-hint", "PasswordHint", "PasswordHint", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:automatic-login:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.AutomaticLogin">"AutomaticLogin"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("automatic-login", "AutomaticLogin", "AutomaticLogin", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:system-account:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.SystemAccount">"SystemAccount"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("system-account", "SystemAccount", "SystemAccount", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:local-account:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.LocalAccount">"LocalAccount"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("local-account", "LocalAccount", "LocalAccount", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * AccountsUser:uses-homed:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.UsesHomed">"UsesHomed"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("uses-homed", "UsesHomed", "UsesHomed", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * accounts_user_get_uid: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Uid">"Uid"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
guint64 
accounts_user_get_uid (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), 0);

  return ACCOUNTS_USER_GET_IFACE (object)->get_uid (object);
}

/**
 * accounts_user_set_uid: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Uid">"Uid"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_uid (AccountsUser *object, guint64 value)
{
  g_object_set (G_OBJECT (object), "uid", value, NULL);
}

/**
 * accounts_user_get_user_name: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.UserName">"UserName"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_user_name() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_user_name (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_user_name (object);
}

/**
 * accounts_user_dup_user_name: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.UserName">"UserName"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_user_name (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "user-name", &value, NULL);
  return value;
}

/**
 * accounts_user_set_user_name: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.UserName">"UserName"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_user_name (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "user-name", value, NULL);
}

/**
 * accounts_user_get_real_name: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.RealName">"RealName"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_real_name() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_real_name (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_real_name (object);
}

/**
 * accounts_user_dup_real_name: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.RealName">"RealName"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_real_name (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "real-name", &value, NULL);
  return value;
}

/**
 * accounts_user_set_real_name: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.RealName">"RealName"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_real_name (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "real-name", value, NULL);
}

/**
 * accounts_user_get_account_type: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.AccountType">"AccountType"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint 
accounts_user_get_account_type (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), 0);

  return ACCOUNTS_USER_GET_IFACE (object)->get_account_type (object);
}

/**
 * accounts_user_set_account_type: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.AccountType">"AccountType"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_account_type (AccountsUser *object, gint value)
{
  g_object_set (G_OBJECT (object), "account-type", value, NULL);
}

/**
 * accounts_user_get_home_directory: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.HomeDirectory">"HomeDirectory"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_home_directory() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_home_directory (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_home_directory (object);
}

/**
 * accounts_user_dup_home_directory: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.HomeDirectory">"HomeDirectory"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_home_directory (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "home-directory", &value, NULL);
  return value;
}

/**
 * accounts_user_set_home_directory: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.HomeDirectory">"HomeDirectory"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_home_directory (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "home-directory", value, NULL);
}

/**
 * accounts_user_get_shell: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Shell">"Shell"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_shell() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_shell (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_shell (object);
}

/**
 * accounts_user_dup_shell: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Shell">"Shell"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_shell (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "shell", &value, NULL);
  return value;
}

/**
 * accounts_user_set_shell: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Shell">"Shell"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_shell (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "shell", value, NULL);
}

/**
 * accounts_user_get_email: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Email">"Email"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_email() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_email (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_email (object);
}

/**
 * accounts_user_dup_email: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Email">"Email"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_email (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "email", &value, NULL);
  return value;
}

/**
 * accounts_user_set_email: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Email">"Email"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_email (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "email", value, NULL);
}

/**
 * accounts_user_get_language: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Language">"Language"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_language() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_language (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_language (object);
}

/**
 * accounts_user_dup_language: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Language">"Language"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_language (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "language", &value, NULL);
  return value;
}

/**
 * accounts_user_set_language: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Language">"Language"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_language (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "language", value, NULL);
}

/**
 * accounts_user_get_languages: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Languages">"Languages"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_languages() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *const *
accounts_user_get_languages (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_languages (object);
}

/**
 * accounts_user_dup_languages: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Languages">"Languages"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_strfreev().
 */
gchar **
accounts_user_dup_languages (AccountsUser *object)
{
  gchar **value;
  g_object_get (G_OBJECT (object), "languages", &value, NULL);
  return value;
}

/**
 * accounts_user_set_languages: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Languages">"Languages"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_languages (AccountsUser *object, const gchar *const *value)
{
  g_object_set (G_OBJECT (object), "languages", value, NULL);
}

/**
 * accounts_user_get_session: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Session">"Session"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_session() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_session (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_session (object);
}

/**
 * accounts_user_dup_session: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Session">"Session"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_session (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "session", &value, NULL);
  return value;
}

/**
 * accounts_user_set_session: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Session">"Session"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_session (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "session", value, NULL);
}

/**
 * accounts_user_get_session_type: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.SessionType">"SessionType"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_session_type() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_session_type (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_session_type (object);
}

/**
 * accounts_user_dup_session_type: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.SessionType">"SessionType"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_session_type (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "session-type", &value, NULL);
  return value;
}

/**
 * accounts_user_set_session_type: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.SessionType">"SessionType"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_session_type (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "session-type", value, NULL);
}

/**
 * accounts_user_get_xsession: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.XSession">"XSession"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_xsession() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_xsession (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_xsession (object);
}

/**
 * accounts_user_dup_xsession: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.XSession">"XSession"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_xsession (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "xsession", &value, NULL);
  return value;
}

/**
 * accounts_user_set_xsession: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.XSession">"XSession"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_xsession (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "xsession", value, NULL);
}

/**
 * accounts_user_get_location: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Location">"Location"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_location() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_location (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_location (object);
}

/**
 * accounts_user_dup_location: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Location">"Location"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_location (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "location", &value, NULL);
  return value;
}

/**
 * accounts_user_set_location: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Location">"Location"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_location (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "location", value, NULL);
}

/**
 * accounts_user_get_login_frequency: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginFrequency">"LoginFrequency"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
guint64 
accounts_user_get_login_frequency (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), 0);

  return ACCOUNTS_USER_GET_IFACE (object)->get_login_frequency (object);
}

/**
 * accounts_user_set_login_frequency: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginFrequency">"LoginFrequency"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_login_frequency (AccountsUser *object, guint64 value)
{
  g_object_set (G_OBJECT (object), "login-frequency", value, NULL);
}

/**
 * accounts_user_get_login_time: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginTime">"LoginTime"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint64 
accounts_user_get_login_time (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), 0);

  return ACCOUNTS_USER_GET_IFACE (object)->get_login_time (object);
}

/**
 * accounts_user_set_login_time: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginTime">"LoginTime"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_login_time (AccountsUser *object, gint64 value)
{
  g_object_set (G_OBJECT (object), "login-time", value, NULL);
}

/**
 * accounts_user_get_login_history: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginHistory">"LoginHistory"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_login_history() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
GVariant *
accounts_user_get_login_history (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_login_history (object);
}

/**
 * accounts_user_dup_login_history: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginHistory">"LoginHistory"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_variant_unref().
 */
GVariant *
accounts_user_dup_login_history (AccountsUser *object)
{
  GVariant *value;
  g_object_get (G_OBJECT (object), "login-history", &value, NULL);
  return value;
}

/**
 * accounts_user_set_login_history: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginHistory">"LoginHistory"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_login_history (AccountsUser *object, GVariant *value)
{
  g_object_set (G_OBJECT (object), "login-history", value, NULL);
}

/**
 * accounts_user_get_icon_file: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.IconFile">"IconFile"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_icon_file() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_icon_file (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_icon_file (object);
}

/**
 * accounts_user_dup_icon_file: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.IconFile">"IconFile"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_icon_file (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "icon-file", &value, NULL);
  return value;
}

/**
 * accounts_user_set_icon_file: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.IconFile">"IconFile"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_icon_file (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "icon-file", value, NULL);
}

/**
 * accounts_user_get_saved: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Saved">"Saved"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
accounts_user_get_saved (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), FALSE);

  return ACCOUNTS_USER_GET_IFACE (object)->get_saved (object);
}

/**
 * accounts_user_set_saved: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Saved">"Saved"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_saved (AccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "saved", value, NULL);
}

/**
 * accounts_user_get_locked: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Locked">"Locked"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
accounts_user_get_locked (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), FALSE);

  return ACCOUNTS_USER_GET_IFACE (object)->get_locked (object);
}

/**
 * accounts_user_set_locked: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Locked">"Locked"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_locked (AccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "locked", value, NULL);
}

/**
 * accounts_user_get_password_mode: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordMode">"PasswordMode"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint 
accounts_user_get_password_mode (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), 0);

  return ACCOUNTS_USER_GET_IFACE (object)->get_password_mode (object);
}

/**
 * accounts_user_set_password_mode: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordMode">"PasswordMode"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_password_mode (AccountsUser *object, gint value)
{
  g_object_set (G_OBJECT (object), "password-mode", value, NULL);
}

/**
 * accounts_user_get_password_hint: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordHint">"PasswordHint"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use accounts_user_dup_password_hint() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
accounts_user_get_password_hint (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), NULL);

  return ACCOUNTS_USER_GET_IFACE (object)->get_password_hint (object);
}

/**
 * accounts_user_dup_password_hint: (skip)
 * @object: A #AccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordHint">"PasswordHint"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
accounts_user_dup_password_hint (AccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "password-hint", &value, NULL);
  return value;
}

/**
 * accounts_user_set_password_hint: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordHint">"PasswordHint"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_password_hint (AccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "password-hint", value, NULL);
}

/**
 * accounts_user_get_automatic_login: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.AutomaticLogin">"AutomaticLogin"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
accounts_user_get_automatic_login (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), FALSE);

  return ACCOUNTS_USER_GET_IFACE (object)->get_automatic_login (object);
}

/**
 * accounts_user_set_automatic_login: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.AutomaticLogin">"AutomaticLogin"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_automatic_login (AccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "automatic-login", value, NULL);
}

/**
 * accounts_user_get_system_account: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.SystemAccount">"SystemAccount"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
accounts_user_get_system_account (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), FALSE);

  return ACCOUNTS_USER_GET_IFACE (object)->get_system_account (object);
}

/**
 * accounts_user_set_system_account: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.SystemAccount">"SystemAccount"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_system_account (AccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "system-account", value, NULL);
}

/**
 * accounts_user_get_local_account: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LocalAccount">"LocalAccount"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
accounts_user_get_local_account (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), FALSE);

  return ACCOUNTS_USER_GET_IFACE (object)->get_local_account (object);
}

/**
 * accounts_user_set_local_account: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LocalAccount">"LocalAccount"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_local_account (AccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "local-account", value, NULL);
}

/**
 * accounts_user_get_uses_homed: (skip)
 * @object: A #AccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.UsesHomed">"UsesHomed"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
accounts_user_get_uses_homed (AccountsUser *object)
{
  g_return_val_if_fail (ACCOUNTS_IS_USER (object), FALSE);

  return ACCOUNTS_USER_GET_IFACE (object)->get_uses_homed (object);
}

/**
 * accounts_user_set_uses_homed: (skip)
 * @object: A #AccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.UsesHomed">"UsesHomed"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
accounts_user_set_uses_homed (AccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "uses-homed", value, NULL);
}

/**
 * accounts_user_emit_changed:
 * @object: A #AccountsUser.
 *
 * Emits the <link linkend="gdbus-signal-org-freedesktop-Accounts-User.Changed">"Changed"</link> D-Bus signal.
 */
void
accounts_user_emit_changed (
    AccountsUser *object)
{
  g_signal_emit (object, ACCOUNTS__USER_SIGNALS[ACCOUNTS__USER_CHANGED], 0);
}

/**
 * accounts_user_call_set_user_name:
 * @proxy: A #AccountsUserProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserName">SetUserName()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_user_name_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_user_name_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_user_name (
    AccountsUser *proxy,
    const gchar *arg_name,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetUserName",
    g_variant_new ("(s)",
                   arg_name),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_user_name_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_user_name().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_user_name().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_user_name_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_user_name_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserName">SetUserName()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_user_name() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_user_name_sync (
    AccountsUser *proxy,
    const gchar *arg_name,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetUserName",
    g_variant_new ("(s)",
                   arg_name),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_real_name:
 * @proxy: A #AccountsUserProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetRealName">SetRealName()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_real_name_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_real_name_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_real_name (
    AccountsUser *proxy,
    const gchar *arg_name,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetRealName",
    g_variant_new ("(s)",
                   arg_name),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_real_name_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_real_name().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_real_name().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_real_name_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_real_name_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetRealName">SetRealName()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_real_name() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_real_name_sync (
    AccountsUser *proxy,
    const gchar *arg_name,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetRealName",
    g_variant_new ("(s)",
                   arg_name),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_email:
 * @proxy: A #AccountsUserProxy.
 * @arg_email: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetEmail">SetEmail()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_email_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_email_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_email (
    AccountsUser *proxy,
    const gchar *arg_email,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetEmail",
    g_variant_new ("(s)",
                   arg_email),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_email_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_email().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_email().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_email_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_email_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_email: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetEmail">SetEmail()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_email() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_email_sync (
    AccountsUser *proxy,
    const gchar *arg_email,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetEmail",
    g_variant_new ("(s)",
                   arg_email),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_language:
 * @proxy: A #AccountsUserProxy.
 * @arg_language: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguage">SetLanguage()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_language_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_language_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_language (
    AccountsUser *proxy,
    const gchar *arg_language,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetLanguage",
    g_variant_new ("(s)",
                   arg_language),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_language_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_language().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_language().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_language_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_language_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_language: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguage">SetLanguage()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_language() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_language_sync (
    AccountsUser *proxy,
    const gchar *arg_language,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetLanguage",
    g_variant_new ("(s)",
                   arg_language),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_languages:
 * @proxy: A #AccountsUserProxy.
 * @arg_languages: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguages">SetLanguages()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_languages_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_languages_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_languages (
    AccountsUser *proxy,
    const gchar *const *arg_languages,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetLanguages",
    g_variant_new ("(^as)",
                   arg_languages),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_languages_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_languages().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_languages().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_languages_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_languages_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_languages: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguages">SetLanguages()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_languages() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_languages_sync (
    AccountsUser *proxy,
    const gchar *const *arg_languages,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetLanguages",
    g_variant_new ("(^as)",
                   arg_languages),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_xsession:
 * @proxy: A #AccountsUserProxy.
 * @arg_x_session: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetXSession">SetXSession()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_xsession_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_xsession_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_xsession (
    AccountsUser *proxy,
    const gchar *arg_x_session,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetXSession",
    g_variant_new ("(s)",
                   arg_x_session),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_xsession_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_xsession().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_xsession().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_xsession_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_xsession_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_x_session: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetXSession">SetXSession()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_xsession() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_xsession_sync (
    AccountsUser *proxy,
    const gchar *arg_x_session,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetXSession",
    g_variant_new ("(s)",
                   arg_x_session),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_session:
 * @proxy: A #AccountsUserProxy.
 * @arg_session: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetSession">SetSession()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_session_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_session_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_session (
    AccountsUser *proxy,
    const gchar *arg_session,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetSession",
    g_variant_new ("(s)",
                   arg_session),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_session_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_session().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_session().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_session_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_session_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_session: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetSession">SetSession()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_session() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_session_sync (
    AccountsUser *proxy,
    const gchar *arg_session,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetSession",
    g_variant_new ("(s)",
                   arg_session),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_session_type:
 * @proxy: A #AccountsUserProxy.
 * @arg_session_type: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetSessionType">SetSessionType()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_session_type_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_session_type_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_session_type (
    AccountsUser *proxy,
    const gchar *arg_session_type,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetSessionType",
    g_variant_new ("(s)",
                   arg_session_type),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_session_type_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_session_type().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_session_type().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_session_type_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_session_type_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_session_type: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetSessionType">SetSessionType()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_session_type() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_session_type_sync (
    AccountsUser *proxy,
    const gchar *arg_session_type,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetSessionType",
    g_variant_new ("(s)",
                   arg_session_type),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_location:
 * @proxy: A #AccountsUserProxy.
 * @arg_location: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocation">SetLocation()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_location_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_location_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_location (
    AccountsUser *proxy,
    const gchar *arg_location,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetLocation",
    g_variant_new ("(s)",
                   arg_location),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_location_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_location().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_location().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_location_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_location_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_location: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocation">SetLocation()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_location() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_location_sync (
    AccountsUser *proxy,
    const gchar *arg_location,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetLocation",
    g_variant_new ("(s)",
                   arg_location),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_home_directory:
 * @proxy: A #AccountsUserProxy.
 * @arg_homedir: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetHomeDirectory">SetHomeDirectory()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_home_directory_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_home_directory_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_home_directory (
    AccountsUser *proxy,
    const gchar *arg_homedir,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetHomeDirectory",
    g_variant_new ("(s)",
                   arg_homedir),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_home_directory_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_home_directory().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_home_directory().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_home_directory_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_home_directory_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_homedir: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetHomeDirectory">SetHomeDirectory()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_home_directory() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_home_directory_sync (
    AccountsUser *proxy,
    const gchar *arg_homedir,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetHomeDirectory",
    g_variant_new ("(s)",
                   arg_homedir),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_shell:
 * @proxy: A #AccountsUserProxy.
 * @arg_shell: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetShell">SetShell()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_shell_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_shell_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_shell (
    AccountsUser *proxy,
    const gchar *arg_shell,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetShell",
    g_variant_new ("(s)",
                   arg_shell),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_shell_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_shell().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_shell().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_shell_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_shell_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_shell: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetShell">SetShell()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_shell() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_shell_sync (
    AccountsUser *proxy,
    const gchar *arg_shell,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetShell",
    g_variant_new ("(s)",
                   arg_shell),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_icon_file:
 * @proxy: A #AccountsUserProxy.
 * @arg_filename: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetIconFile">SetIconFile()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_icon_file_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_icon_file_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_icon_file (
    AccountsUser *proxy,
    const gchar *arg_filename,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetIconFile",
    g_variant_new ("(s)",
                   arg_filename),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_icon_file_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_icon_file().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_icon_file().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_icon_file_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_icon_file_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_filename: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetIconFile">SetIconFile()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_icon_file() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_icon_file_sync (
    AccountsUser *proxy,
    const gchar *arg_filename,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetIconFile",
    g_variant_new ("(s)",
                   arg_filename),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_locked:
 * @proxy: A #AccountsUserProxy.
 * @arg_locked: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocked">SetLocked()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_locked_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_locked_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_locked (
    AccountsUser *proxy,
    gboolean arg_locked,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetLocked",
    g_variant_new ("(b)",
                   arg_locked),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_locked_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_locked().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_locked().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_locked_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_locked_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_locked: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocked">SetLocked()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_locked() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_locked_sync (
    AccountsUser *proxy,
    gboolean arg_locked,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetLocked",
    g_variant_new ("(b)",
                   arg_locked),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_account_type:
 * @proxy: A #AccountsUserProxy.
 * @arg_accountType: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAccountType">SetAccountType()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_account_type_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_account_type_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_account_type (
    AccountsUser *proxy,
    gint arg_accountType,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetAccountType",
    g_variant_new ("(i)",
                   arg_accountType),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_account_type_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_account_type().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_account_type().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_account_type_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_account_type_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_accountType: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAccountType">SetAccountType()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_account_type() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_account_type_sync (
    AccountsUser *proxy,
    gint arg_accountType,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetAccountType",
    g_variant_new ("(i)",
                   arg_accountType),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_password_mode:
 * @proxy: A #AccountsUserProxy.
 * @arg_mode: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordMode">SetPasswordMode()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_password_mode_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_password_mode_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_password_mode (
    AccountsUser *proxy,
    gint arg_mode,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetPasswordMode",
    g_variant_new ("(i)",
                   arg_mode),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_password_mode_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_password_mode().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_password_mode().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_password_mode_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_password_mode_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_mode: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordMode">SetPasswordMode()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_password_mode() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_password_mode_sync (
    AccountsUser *proxy,
    gint arg_mode,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetPasswordMode",
    g_variant_new ("(i)",
                   arg_mode),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_password:
 * @proxy: A #AccountsUserProxy.
 * @arg_password: Argument to pass with the method invocation.
 * @arg_hint: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPassword">SetPassword()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_password_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_password_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_password (
    AccountsUser *proxy,
    const gchar *arg_password,
    const gchar *arg_hint,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetPassword",
    g_variant_new ("(ss)",
                   arg_password,
                   arg_hint),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_password_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_password().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_password().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_password_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_password_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_password: Argument to pass with the method invocation.
 * @arg_hint: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPassword">SetPassword()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_password() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_password_sync (
    AccountsUser *proxy,
    const gchar *arg_password,
    const gchar *arg_hint,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetPassword",
    g_variant_new ("(ss)",
                   arg_password,
                   arg_hint),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_password_hint:
 * @proxy: A #AccountsUserProxy.
 * @arg_hint: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordHint">SetPasswordHint()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_password_hint_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_password_hint_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_password_hint (
    AccountsUser *proxy,
    const gchar *arg_hint,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetPasswordHint",
    g_variant_new ("(s)",
                   arg_hint),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_password_hint_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_password_hint().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_password_hint().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_password_hint_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_password_hint_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_hint: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordHint">SetPasswordHint()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_password_hint() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_password_hint_sync (
    AccountsUser *proxy,
    const gchar *arg_hint,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetPasswordHint",
    g_variant_new ("(s)",
                   arg_hint),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_automatic_login:
 * @proxy: A #AccountsUserProxy.
 * @arg_enabled: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAutomaticLogin">SetAutomaticLogin()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_automatic_login_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_automatic_login_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_automatic_login (
    AccountsUser *proxy,
    gboolean arg_enabled,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetAutomaticLogin",
    g_variant_new ("(b)",
                   arg_enabled),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_automatic_login_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_automatic_login().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_automatic_login().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_automatic_login_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_automatic_login_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_enabled: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAutomaticLogin">SetAutomaticLogin()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_automatic_login() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_automatic_login_sync (
    AccountsUser *proxy,
    gboolean arg_enabled,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetAutomaticLogin",
    g_variant_new ("(b)",
                   arg_enabled),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_get_password_expiration_policy:
 * @proxy: A #AccountsUserProxy.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.GetPasswordExpirationPolicy">GetPasswordExpirationPolicy()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_get_password_expiration_policy_finish() to get the result of the operation.
 *
 * See accounts_user_call_get_password_expiration_policy_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_get_password_expiration_policy (
    AccountsUser *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GetPasswordExpirationPolicy",
    g_variant_new ("()"),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_get_password_expiration_policy_finish:
 * @proxy: A #AccountsUserProxy.
 * @out_expiration_time: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_last_change_time: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_min_days_between_changes: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_max_days_between_changes: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_days_to_warn: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_days_after_expiration_until_lock: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_get_password_expiration_policy().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_get_password_expiration_policy().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_get_password_expiration_policy_finish (
    AccountsUser *proxy,
    gint64* out_expiration_time,
    gint64* out_last_change_time,
    gint64* out_min_days_between_changes,
    gint64* out_max_days_between_changes,
    gint64* out_days_to_warn,
    gint64* out_days_after_expiration_until_lock,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(xxxxxx)",
                 out_expiration_time,
                 out_last_change_time,
                 out_min_days_between_changes,
                 out_max_days_between_changes,
                 out_days_to_warn,
                 out_days_after_expiration_until_lock);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_get_password_expiration_policy_sync:
 * @proxy: A #AccountsUserProxy.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @out_expiration_time: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_last_change_time: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_min_days_between_changes: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_max_days_between_changes: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_days_to_warn: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_days_after_expiration_until_lock: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.GetPasswordExpirationPolicy">GetPasswordExpirationPolicy()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_get_password_expiration_policy() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_get_password_expiration_policy_sync (
    AccountsUser *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gint64* out_expiration_time,
    gint64* out_last_change_time,
    gint64* out_min_days_between_changes,
    gint64* out_max_days_between_changes,
    gint64* out_days_to_warn,
    gint64* out_days_after_expiration_until_lock,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GetPasswordExpirationPolicy",
    g_variant_new ("()"),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(xxxxxx)",
                 out_expiration_time,
                 out_last_change_time,
                 out_min_days_between_changes,
                 out_max_days_between_changes,
                 out_days_to_warn,
                 out_days_after_expiration_until_lock);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_password_expiration_policy:
 * @proxy: A #AccountsUserProxy.
 * @arg_min_days_between_changes: Argument to pass with the method invocation.
 * @arg_max_days_between_changes: Argument to pass with the method invocation.
 * @arg_days_to_warn: Argument to pass with the method invocation.
 * @arg_days_after_expiration_until_lock: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordExpirationPolicy">SetPasswordExpirationPolicy()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_password_expiration_policy_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_password_expiration_policy_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_password_expiration_policy (
    AccountsUser *proxy,
    gint64 arg_min_days_between_changes,
    gint64 arg_max_days_between_changes,
    gint64 arg_days_to_warn,
    gint64 arg_days_after_expiration_until_lock,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetPasswordExpirationPolicy",
    g_variant_new ("(xxxx)",
                   arg_min_days_between_changes,
                   arg_max_days_between_changes,
                   arg_days_to_warn,
                   arg_days_after_expiration_until_lock),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_password_expiration_policy_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_password_expiration_policy().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_password_expiration_policy().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_password_expiration_policy_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_password_expiration_policy_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_min_days_between_changes: Argument to pass with the method invocation.
 * @arg_max_days_between_changes: Argument to pass with the method invocation.
 * @arg_days_to_warn: Argument to pass with the method invocation.
 * @arg_days_after_expiration_until_lock: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordExpirationPolicy">SetPasswordExpirationPolicy()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_password_expiration_policy() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_password_expiration_policy_sync (
    AccountsUser *proxy,
    gint64 arg_min_days_between_changes,
    gint64 arg_max_days_between_changes,
    gint64 arg_days_to_warn,
    gint64 arg_days_after_expiration_until_lock,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetPasswordExpirationPolicy",
    g_variant_new ("(xxxx)",
                   arg_min_days_between_changes,
                   arg_max_days_between_changes,
                   arg_days_to_warn,
                   arg_days_after_expiration_until_lock),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_user_expiration_policy:
 * @proxy: A #AccountsUserProxy.
 * @arg_expiration_time: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserExpirationPolicy">SetUserExpirationPolicy()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_call_set_user_expiration_policy_finish() to get the result of the operation.
 *
 * See accounts_user_call_set_user_expiration_policy_sync() for the synchronous, blocking version of this method.
 */
void
accounts_user_call_set_user_expiration_policy (
    AccountsUser *proxy,
    gint64 arg_expiration_time,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetUserExpirationPolicy",
    g_variant_new ("(x)",
                   arg_expiration_time),
    call_flags,
    timeout_msec,
    cancellable,
    callback,
    user_data);
}

/**
 * accounts_user_call_set_user_expiration_policy_finish:
 * @proxy: A #AccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_call_set_user_expiration_policy().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with accounts_user_call_set_user_expiration_policy().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_user_expiration_policy_finish (
    AccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_call_set_user_expiration_policy_sync:
 * @proxy: A #AccountsUserProxy.
 * @arg_expiration_time: Argument to pass with the method invocation.
 * @call_flags: Flags from the #GDBusCallFlags enumeration. If you want to allow interactive
       authorization be sure to set %G_DBUS_CALL_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION.
 * @timeout_msec: The timeout in milliseconds (with %G_MAXINT meaning "infinite") or
       -1 to use the proxy default timeout.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserExpirationPolicy">SetUserExpirationPolicy()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See accounts_user_call_set_user_expiration_policy() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
accounts_user_call_set_user_expiration_policy_sync (
    AccountsUser *proxy,
    gint64 arg_expiration_time,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetUserExpirationPolicy",
    g_variant_new ("(x)",
                   arg_expiration_time),
    call_flags,
    timeout_msec,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * accounts_user_complete_set_user_name:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserName">SetUserName()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_user_name (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_real_name:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetRealName">SetRealName()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_real_name (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_email:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetEmail">SetEmail()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_email (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_language:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguage">SetLanguage()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_language (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_languages:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguages">SetLanguages()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_languages (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_xsession:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetXSession">SetXSession()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_xsession (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_session:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetSession">SetSession()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_session (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_session_type:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetSessionType">SetSessionType()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_session_type (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_location:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocation">SetLocation()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_location (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_home_directory:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetHomeDirectory">SetHomeDirectory()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_home_directory (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_shell:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetShell">SetShell()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_shell (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_icon_file:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetIconFile">SetIconFile()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_icon_file (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_locked:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocked">SetLocked()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_locked (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_account_type:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAccountType">SetAccountType()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_account_type (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_password_mode:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordMode">SetPasswordMode()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_password_mode (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_password:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPassword">SetPassword()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_password (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_password_hint:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordHint">SetPasswordHint()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_password_hint (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_automatic_login:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAutomaticLogin">SetAutomaticLogin()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_automatic_login (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_get_password_expiration_policy:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @expiration_time: Parameter to return.
 * @last_change_time: Parameter to return.
 * @min_days_between_changes: Parameter to return.
 * @max_days_between_changes: Parameter to return.
 * @days_to_warn: Parameter to return.
 * @days_after_expiration_until_lock: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.GetPasswordExpirationPolicy">GetPasswordExpirationPolicy()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_get_password_expiration_policy (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gint64 expiration_time,
    gint64 last_change_time,
    gint64 min_days_between_changes,
    gint64 max_days_between_changes,
    gint64 days_to_warn,
    gint64 days_after_expiration_until_lock)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(xxxxxx)",
                   expiration_time,
                   last_change_time,
                   min_days_between_changes,
                   max_days_between_changes,
                   days_to_warn,
                   days_after_expiration_until_lock));
}

/**
 * accounts_user_complete_set_password_expiration_policy:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordExpirationPolicy">SetPasswordExpirationPolicy()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_password_expiration_policy (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * accounts_user_complete_set_user_expiration_policy:
 * @object: A #AccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserExpirationPolicy">SetUserExpirationPolicy()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
accounts_user_complete_set_user_expiration_policy (
    AccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * AccountsUserProxy:
 *
 * The #AccountsUserProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * AccountsUserProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #AccountsUserProxy.
 */

struct _AccountsUserProxyPrivate
{
  GData *qdata;
};

static void accounts_user_proxy_iface_init (AccountsUserIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (AccountsUserProxy, accounts_user_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (AccountsUserProxy)
                         G_IMPLEMENT_INTERFACE (ACCOUNTS_TYPE_USER, accounts_user_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (AccountsUserProxy, accounts_user_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (ACCOUNTS_TYPE_USER, accounts_user_proxy_iface_init))

#endif
static void
accounts_user_proxy_finalize (GObject *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (accounts_user_proxy_parent_class)->finalize (object);
}

static void
accounts_user_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 25);
  info = (const _ExtendedGDBusPropertyInfo *) _accounts_user_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
accounts_user_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.freedesktop.Accounts.User: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
accounts_user_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 25);
  info = (const _ExtendedGDBusPropertyInfo *) _accounts_user_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.freedesktop.Accounts.User", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) accounts_user_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
accounts_user_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_accounts_user_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], ACCOUNTS_TYPE_USER);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, ACCOUNTS_TYPE_USER);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
accounts_user_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_accounts_user_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_accounts_user_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static guint64 
accounts_user_proxy_get_uid (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  guint64 value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Uid");
  if (variant != NULL)
    {
      value = g_variant_get_uint64 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_user_name (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "UserName");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_real_name (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "RealName");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static gint 
accounts_user_proxy_get_account_type (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "AccountType");
  if (variant != NULL)
    {
      value = g_variant_get_int32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_home_directory (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "HomeDirectory");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_shell (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Shell");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_email (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Email");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_language (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Language");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *const *
accounts_user_proxy_get_languages (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *const *value = NULL;
  value = g_datalist_get_data (&proxy->priv->qdata, "Languages");
  if (value != NULL)
    return value;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Languages");
  if (variant != NULL)
    {
      value = g_variant_get_strv (variant, NULL);
      g_datalist_set_data_full (&proxy->priv->qdata, "Languages", (gpointer) value, g_free);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_session (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Session");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_session_type (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "SessionType");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_xsession (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "XSession");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_location (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Location");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static guint64 
accounts_user_proxy_get_login_frequency (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  guint64 value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "LoginFrequency");
  if (variant != NULL)
    {
      value = g_variant_get_uint64 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gint64 
accounts_user_proxy_get_login_time (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gint64 value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "LoginTime");
  if (variant != NULL)
    {
      value = g_variant_get_int64 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static GVariant *
accounts_user_proxy_get_login_history (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  GVariant *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "LoginHistory");
  value = variant;
  if (variant != NULL)
    g_variant_unref (variant);
  return value;
}

static const gchar *
accounts_user_proxy_get_icon_file (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "IconFile");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
accounts_user_proxy_get_saved (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Saved");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
accounts_user_proxy_get_locked (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Locked");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gint 
accounts_user_proxy_get_password_mode (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "PasswordMode");
  if (variant != NULL)
    {
      value = g_variant_get_int32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
accounts_user_proxy_get_password_hint (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "PasswordHint");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
accounts_user_proxy_get_automatic_login (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "AutomaticLogin");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
accounts_user_proxy_get_system_account (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "SystemAccount");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
accounts_user_proxy_get_local_account (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "LocalAccount");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
accounts_user_proxy_get_uses_homed (AccountsUser *object)
{
  AccountsUserProxy *proxy = ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "UsesHomed");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static void
accounts_user_proxy_init (AccountsUserProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = accounts_user_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, ACCOUNTS_TYPE_USER_PROXY, AccountsUserProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), accounts_user_interface_info ());
}

static void
accounts_user_proxy_class_init (AccountsUserProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = accounts_user_proxy_finalize;
  gobject_class->get_property = accounts_user_proxy_get_property;
  gobject_class->set_property = accounts_user_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = accounts_user_proxy_g_signal;
  proxy_class->g_properties_changed = accounts_user_proxy_g_properties_changed;

  accounts_user_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (AccountsUserProxyPrivate));
#endif
}

static void
accounts_user_proxy_iface_init (AccountsUserIface *iface)
{
  iface->get_uid = accounts_user_proxy_get_uid;
  iface->get_user_name = accounts_user_proxy_get_user_name;
  iface->get_real_name = accounts_user_proxy_get_real_name;
  iface->get_account_type = accounts_user_proxy_get_account_type;
  iface->get_home_directory = accounts_user_proxy_get_home_directory;
  iface->get_shell = accounts_user_proxy_get_shell;
  iface->get_email = accounts_user_proxy_get_email;
  iface->get_language = accounts_user_proxy_get_language;
  iface->get_languages = accounts_user_proxy_get_languages;
  iface->get_session = accounts_user_proxy_get_session;
  iface->get_session_type = accounts_user_proxy_get_session_type;
  iface->get_xsession = accounts_user_proxy_get_xsession;
  iface->get_location = accounts_user_proxy_get_location;
  iface->get_login_frequency = accounts_user_proxy_get_login_frequency;
  iface->get_login_time = accounts_user_proxy_get_login_time;
  iface->get_login_history = accounts_user_proxy_get_login_history;
  iface->get_icon_file = accounts_user_proxy_get_icon_file;
  iface->get_saved = accounts_user_proxy_get_saved;
  iface->get_locked = accounts_user_proxy_get_locked;
  iface->get_password_mode = accounts_user_proxy_get_password_mode;
  iface->get_password_hint = accounts_user_proxy_get_password_hint;
  iface->get_automatic_login = accounts_user_proxy_get_automatic_login;
  iface->get_system_account = accounts_user_proxy_get_system_account;
  iface->get_local_account = accounts_user_proxy_get_local_account;
  iface->get_uses_homed = accounts_user_proxy_get_uses_homed;
}

/**
 * accounts_user_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_proxy_new_finish() to get the result of the operation.
 *
 * See accounts_user_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
accounts_user_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (ACCOUNTS_TYPE_USER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts.User", NULL);
}

/**
 * accounts_user_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with accounts_user_proxy_new().
 *
 * Returns: (transfer full) (type AccountsUserProxy): The constructed proxy object or %NULL if @error is set.
 */
AccountsUser *
accounts_user_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ACCOUNTS_USER (ret);
  else
    return NULL;
}

/**
 * accounts_user_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See accounts_user_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type AccountsUserProxy): The constructed proxy object or %NULL if @error is set.
 */
AccountsUser *
accounts_user_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (ACCOUNTS_TYPE_USER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts.User", NULL);
  if (ret != NULL)
    return ACCOUNTS_USER (ret);
  else
    return NULL;
}


/**
 * accounts_user_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like accounts_user_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call accounts_user_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See accounts_user_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
accounts_user_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (ACCOUNTS_TYPE_USER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts.User", NULL);
}

/**
 * accounts_user_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to accounts_user_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with accounts_user_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type AccountsUserProxy): The constructed proxy object or %NULL if @error is set.
 */
AccountsUser *
accounts_user_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ACCOUNTS_USER (ret);
  else
    return NULL;
}

/**
 * accounts_user_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like accounts_user_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See accounts_user_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type AccountsUserProxy): The constructed proxy object or %NULL if @error is set.
 */
AccountsUser *
accounts_user_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (ACCOUNTS_TYPE_USER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts.User", NULL);
  if (ret != NULL)
    return ACCOUNTS_USER (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * AccountsUserSkeleton:
 *
 * The #AccountsUserSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * AccountsUserSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #AccountsUserSkeleton.
 */

struct _AccountsUserSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_accounts_user_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], ACCOUNTS_TYPE_USER);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, ACCOUNTS_TYPE_USER);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_accounts_user_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_accounts_user_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_accounts_user_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_accounts_user_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _accounts_user_skeleton_vtable =
{
  _accounts_user_skeleton_handle_method_call,
  _accounts_user_skeleton_handle_get_property,
  _accounts_user_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
accounts_user_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return accounts_user_interface_info ();
}

static GDBusInterfaceVTable *
accounts_user_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_accounts_user_skeleton_vtable;
}

static GVariant *
accounts_user_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_accounts_user_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _accounts_user_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _accounts_user_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _accounts_user_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.freedesktop.Accounts.User", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _accounts_user_emit_changed (gpointer user_data);

static void
accounts_user_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _accounts_user_emit_changed (skeleton);
}

static void
_accounts_user_on_signal_changed (
    AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.freedesktop.Accounts.User", "Changed",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void accounts_user_skeleton_iface_init (AccountsUserIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (AccountsUserSkeleton, accounts_user_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (AccountsUserSkeleton)
                         G_IMPLEMENT_INTERFACE (ACCOUNTS_TYPE_USER, accounts_user_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (AccountsUserSkeleton, accounts_user_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (ACCOUNTS_TYPE_USER, accounts_user_skeleton_iface_init))

#endif
static void
accounts_user_skeleton_finalize (GObject *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  guint n;
  for (n = 0; n < 25; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (accounts_user_skeleton_parent_class)->finalize (object);
}

static void
accounts_user_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 25);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_accounts_user_emit_changed (gpointer user_data)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.freedesktop.Accounts.User",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_accounts_user_schedule_emit_changed (AccountsUserSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
accounts_user_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _accounts_user_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _accounts_user_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
accounts_user_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 25);
  info = (const _ExtendedGDBusPropertyInfo *) _accounts_user_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _accounts_user_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
accounts_user_skeleton_init (AccountsUserSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = accounts_user_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, ACCOUNTS_TYPE_USER_SKELETON, AccountsUserSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 25);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_UINT64);
  g_value_init (&skeleton->priv->properties[1], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[2], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[3], G_TYPE_INT);
  g_value_init (&skeleton->priv->properties[4], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[5], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[6], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[7], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[8], G_TYPE_STRV);
  g_value_init (&skeleton->priv->properties[9], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[10], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[11], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[12], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[13], G_TYPE_UINT64);
  g_value_init (&skeleton->priv->properties[14], G_TYPE_INT64);
  g_value_init (&skeleton->priv->properties[15], G_TYPE_VARIANT);
  g_value_init (&skeleton->priv->properties[16], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[17], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[18], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[19], G_TYPE_INT);
  g_value_init (&skeleton->priv->properties[20], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[21], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[22], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[23], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[24], G_TYPE_BOOLEAN);
}

static guint64 
accounts_user_skeleton_get_uid (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  guint64 value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_uint64 (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_user_name (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[1]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_real_name (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[2]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gint 
accounts_user_skeleton_get_account_type (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int (&(skeleton->priv->properties[3]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_home_directory (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[4]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_shell (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[5]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_email (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[6]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_language (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[7]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *const *
accounts_user_skeleton_get_languages (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *const *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boxed (&(skeleton->priv->properties[8]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_session (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[9]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_session_type (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[10]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_xsession (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[11]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_location (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[12]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static guint64 
accounts_user_skeleton_get_login_frequency (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  guint64 value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_uint64 (&(skeleton->priv->properties[13]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gint64 
accounts_user_skeleton_get_login_time (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gint64 value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int64 (&(skeleton->priv->properties[14]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static GVariant *
accounts_user_skeleton_get_login_history (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  GVariant *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_variant (&(skeleton->priv->properties[15]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_icon_file (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[16]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
accounts_user_skeleton_get_saved (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[17]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
accounts_user_skeleton_get_locked (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[18]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gint 
accounts_user_skeleton_get_password_mode (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int (&(skeleton->priv->properties[19]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
accounts_user_skeleton_get_password_hint (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[20]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
accounts_user_skeleton_get_automatic_login (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[21]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
accounts_user_skeleton_get_system_account (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[22]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
accounts_user_skeleton_get_local_account (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[23]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
accounts_user_skeleton_get_uses_homed (AccountsUser *object)
{
  AccountsUserSkeleton *skeleton = ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[24]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
accounts_user_skeleton_class_init (AccountsUserSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = accounts_user_skeleton_finalize;
  gobject_class->get_property = accounts_user_skeleton_get_property;
  gobject_class->set_property = accounts_user_skeleton_set_property;
  gobject_class->notify       = accounts_user_skeleton_notify;


  accounts_user_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = accounts_user_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = accounts_user_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = accounts_user_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = accounts_user_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (AccountsUserSkeletonPrivate));
#endif
}

static void
accounts_user_skeleton_iface_init (AccountsUserIface *iface)
{
  iface->changed = _accounts_user_on_signal_changed;
  iface->get_uid = accounts_user_skeleton_get_uid;
  iface->get_user_name = accounts_user_skeleton_get_user_name;
  iface->get_real_name = accounts_user_skeleton_get_real_name;
  iface->get_account_type = accounts_user_skeleton_get_account_type;
  iface->get_home_directory = accounts_user_skeleton_get_home_directory;
  iface->get_shell = accounts_user_skeleton_get_shell;
  iface->get_email = accounts_user_skeleton_get_email;
  iface->get_language = accounts_user_skeleton_get_language;
  iface->get_languages = accounts_user_skeleton_get_languages;
  iface->get_session = accounts_user_skeleton_get_session;
  iface->get_session_type = accounts_user_skeleton_get_session_type;
  iface->get_xsession = accounts_user_skeleton_get_xsession;
  iface->get_location = accounts_user_skeleton_get_location;
  iface->get_login_frequency = accounts_user_skeleton_get_login_frequency;
  iface->get_login_time = accounts_user_skeleton_get_login_time;
  iface->get_login_history = accounts_user_skeleton_get_login_history;
  iface->get_icon_file = accounts_user_skeleton_get_icon_file;
  iface->get_saved = accounts_user_skeleton_get_saved;
  iface->get_locked = accounts_user_skeleton_get_locked;
  iface->get_password_mode = accounts_user_skeleton_get_password_mode;
  iface->get_password_hint = accounts_user_skeleton_get_password_hint;
  iface->get_automatic_login = accounts_user_skeleton_get_automatic_login;
  iface->get_system_account = accounts_user_skeleton_get_system_account;
  iface->get_local_account = accounts_user_skeleton_get_local_account;
  iface->get_uses_homed = accounts_user_skeleton_get_uses_homed;
}

/**
 * accounts_user_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>.
 *
 * Returns: (transfer full) (type AccountsUserSkeleton): The skeleton object.
 */
AccountsUser *
accounts_user_skeleton_new (void)
{
  return ACCOUNTS_USER (g_object_new (ACCOUNTS_TYPE_USER_SKELETON, NULL));
}

